from cg.constants.orderforms import REV_SEX_MAP, SOURCE_TYPES
from cg.models.orders.sample_base import PriorityEnum


def parse_panels(panels: str) -> list[str] | None:
    if not panels:
        return None
    separator = ";" if ";" in panels else None
    if ":" in panels:
        separator = ":"
    return panels.split(separator)


def validate_data_analysis(data_analysis):
    data_analysis_alternatives = [
        "Balsamic",  # OF 1508
        "Balsamic QC",  # OF 1508
        "Balsamic UMI",  # OF 1508
        "fastq",  # OF 1605
        "FLUFFY",  # OF 1604
        "MicroSALT",  # OF 1603 (implicit)
        "MIP DNA",  # OF 1508
        "MIP RNA",  # OF 1508
        "RNAfusion",  # OF 1508
        "SARS-CoV-2",  # OF 2184
        "No analysis",  # OF 1508, 1604, 2184
    ]
    if data_analysis not in data_analysis_alternatives:
        raise AttributeError(f"'{data_analysis}' is not a valid data analysis")
    return data_analysis


def numeric_value(value: str | None) -> str | None:
    """Validates that the given string can be given as either an integer or a float. Also converts floats of the
    type x.00 to x."""
    if not value:
        return None
    try:
        float_value: float = float(value)
        if float_value.is_integer():
            return str(int(float_value))
        return value
    except ValueError:
        raise AttributeError(f"Order contains non-numeric value '{value}'")


def validate_parent(parent: str) -> str:
    return None if parent == "0.0" else parent


def validate_source(source: str | None) -> str:
    if source not in SOURCE_TYPES:
        raise ValueError(f"'{source}' is not a valid source")
    return source


def convert_sex(sex: str) -> str | None:
    if not sex:
        return None
    sex = sex.strip()
    return REV_SEX_MAP.get(sex, "unknown")


def convert_to_lower(value: str) -> str:
    return value.lower()


def replace_spaces_with_underscores(value: str) -> str:
    return value.replace(" ", "_")


def convert_to_priority(priority: str | None) -> str | None:
    """Translates the Swedish 'förtur' to 'priority' if specified in the order."""
    return PriorityEnum.priority if priority == "förtur" else priority


def convert_to_date(date: str | None) -> str | None:
    return date[:10] if date else None



from pydantic.v1 import BaseModel, validator

from cg.models.deliverables.metric_deliverables import MetricCondition, MetricsBase


def percent_value_validation(cls, value: float) -> float:
    """Converts a raw float value to percent"""

    return value * 100


class BalsamicMetricsBase(MetricsBase):
    """BALSAMIC base metric attributes

    Attributes:
        condition: balsamic metric validation condition
    """

    condition: MetricCondition | None


class BalsamicQCMetrics(BaseModel):
    """BALSAMIC common QC metrics"""

    mean_insert_size: float | None
    fold_80_base_penalty: float | None


class BalsamicTargetedQCMetrics(BalsamicQCMetrics):
    """BALSAMIC targeted QC metrics"""

    mean_target_coverage: float | None
    median_target_coverage: float | None
    percent_duplication: float | None
    pct_target_bases_50x: float | None
    pct_target_bases_100x: float | None
    pct_target_bases_250x: float | None
    pct_target_bases_500x: float | None
    pct_target_bases_1000x: float | None
    pct_off_bait: float | None
    gc_dropout: float | None

    _pct_values = validator(
        "percent_duplication",
        "pct_target_bases_50x",
        "pct_target_bases_100x",
        "pct_target_bases_250x",
        "pct_target_bases_500x",
        "pct_target_bases_1000x",
        "pct_off_bait",
        allow_reuse=True,
    )(percent_value_validation)


class BalsamicWGSQCMetrics(BalsamicQCMetrics):
    """BALSAMIC WGS QC metrics"""

    median_coverage: float | None
    percent_duplication_r1: float | None
    percent_duplication_r2: float | None
    pct_15x: float | None
    pct_30x: float | None
    pct_60x: float | None
    pct_100x: float | None
    pct_pf_reads_improper_pairs: float | None

    _pct_values = validator(
        "pct_15x",
        "pct_30x",
        "pct_60x",
        "pct_100x",
        "pct_pf_reads_improper_pairs",
        allow_reuse=True,
    )(percent_value_validation)



import shutil
from pathlib import Path

from cg.cli.demultiplex.demux import create_manifest_files
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.io.csv import read_csv
from tests.meta.demultiplex.conftest import get_all_files_in_directory_tree


def test_create_manifest_files_true(
    cli_runner, tmp_flow_cells_directory_ready_for_demultiplexing_bcl_convert: Path
):
    # GIVEN two flowcell directories
    first_flowcell_directory: Path = tmp_flow_cells_directory_ready_for_demultiplexing_bcl_convert
    second_flowcell_directory = Path(
        shutil.copytree(
            first_flowcell_directory,
            Path(
                tmp_flow_cells_directory_ready_for_demultiplexing_bcl_convert.parent, "flowcell_2"
            ),
        )
    )

    # WHEN creating manifest files
    cli_runner.invoke(
        cli=create_manifest_files,
        args=["--source-directory", first_flowcell_directory.parent.as_posix()],
    )

    for flow_cell in [first_flowcell_directory, second_flowcell_directory]:
        manifest_file = Path(flow_cell, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)

        # THEN the manifest file should exist
        assert manifest_file.exists()

        all_files_in_directory: list[Path] = get_all_files_in_directory_tree(flow_cell)
        all_files_in_directory.remove(Path(manifest_file.name))
        files_in_manifest: list[Path] = [
            Path(file[0].strip()) for file in read_csv(delimiter="\t", file_path=manifest_file)
        ]

        # THEN the manifest file should contain all files in the flowcell directory
        for file in all_files_in_directory:
            assert file in files_in_manifest


def test_create_manifest_files_false(
    cli_runner, tmp_flow_cells_directory_ready_for_demultiplexing_bcl_convert: Path
):
    # GIVEN two flowcell directories with missing CopyComplete.txt
    first_flowcell_directory = tmp_flow_cells_directory_ready_for_demultiplexing_bcl_convert
    Path(first_flowcell_directory, DemultiplexingDirsAndFiles.COPY_COMPLETE).unlink()
    second_flowcell_directory = Path(
        shutil.copytree(
            first_flowcell_directory,
            Path(
                tmp_flow_cells_directory_ready_for_demultiplexing_bcl_convert.parent, "flowcell_2"
            ),
        )
    )

    # WHEN creating manifest files
    cli_runner.invoke(
        cli=create_manifest_files,
        args=["--source-directory", first_flowcell_directory.parent.as_posix()],
    )

    for flow_cell in [first_flowcell_directory, second_flowcell_directory]:
        manifest_file = Path(flow_cell, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)

        # THEN the manifest file should not exist
        assert not manifest_file.exists()



"""Model MIP config"""


from pydantic.v1 import BaseModel, EmailStr, Field, validator

from cg.constants.priority import SlurmQos


class AnalysisType(BaseModel):
    analysis_type: str
    sample_id: str


class MipBaseConfig(BaseModel):
    """This model is used when validating the mip analysis config"""

    family_id_: str = Field(None, alias="family_id")
    case_id: str = None
    analysis_type_: dict = Field(..., alias="analysis_type")
    samples: list[AnalysisType] = None
    config_path: str = Field(..., alias="config_file_analysis")
    deliverables_file_path: str = Field(..., alias="store_file")
    email: EmailStr
    is_dry_run: bool = Field(False, alias="dry_run_all")
    log_path: str = Field(..., alias="log_file")
    out_dir: str = Field(..., alias="outdata_dir")
    priority: SlurmQos = Field(..., alias="slurm_quality_of_service")
    sample_info_path: str = Field(..., alias="sample_info_file")
    sample_ids: list[str]

    @validator("case_id", always=True, pre=True)
    def set_case_id(cls, value, values: dict) -> str:
        """Set case_id. Family_id is used for older versions of MIP analysis"""
        return value or values.get("family_id_")

    @validator("samples", always=True, pre=True)
    def set_samples(cls, _, values: dict) -> list[AnalysisType]:
        """Set samples analysis type"""
        raw_samples: dict = values.get("analysis_type_")
        return [
            AnalysisType(sample_id=sample_id, analysis_type=analysis_type)
            for sample_id, analysis_type in raw_samples.items()
        ]



"""RNAfusion delivery report API."""


from cg.constants import (
    REQUIRED_APPLICATION_FIELDS,
    REQUIRED_CASE_FIELDS,
    REQUIRED_CUSTOMER_FIELDS,
    REQUIRED_DATA_ANALYSIS_RNAFUSION_FIELDS,
    REQUIRED_REPORT_FIELDS,
    REQUIRED_SAMPLE_METADATA_RNAFUSION_FIELDS,
    REQUIRED_SAMPLE_METHODS_FIELDS,
    REQUIRED_SAMPLE_RNAFUSION_FIELDS,
    REQUIRED_SAMPLE_TIMESTAMP_FIELDS,
    Pipeline,
)
from cg.constants.constants import GenomeVersion
from cg.meta.report.field_validators import get_million_read_pairs
from cg.meta.report.report_api import ReportAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.analysis import AnalysisModel
from cg.models.cg_config import CGConfig
from cg.models.report.metadata import RnafusionSampleMetadataModel
from cg.models.report.report import CaseModel
from cg.models.report.sample import SampleModel
from cg.models.rnafusion.rnafusion import RnafusionAnalysis, RnafusionQCMetrics
from cg.store.models import Case, Sample


class RnafusionReportAPI(ReportAPI):
    """API to create RNAfusion delivery reports."""

    def __init__(self, config: CGConfig, analysis_api: RnafusionAnalysisAPI):
        super().__init__(config=config, analysis_api=analysis_api)
        self.analysis_api: RnafusionAnalysisAPI = analysis_api

    def get_sample_metadata(
        self, case: Case, sample: Sample, analysis_metadata: RnafusionAnalysis
    ) -> RnafusionSampleMetadataModel:
        """Return sample metadata to include in the report."""
        sample_metrics: RnafusionQCMetrics = analysis_metadata.sample_metrics[sample.internal_id]

        # Skip LIMS data collection if down sampled
        input_amount = None
        rin = None
        if not sample.downsampled_to:
            input_amount = self.lims_api.get_latest_rna_input_amount(sample_id=sample.internal_id)
            rin = self.lims_api.get_sample_rin(sample_id=sample.internal_id)

        return RnafusionSampleMetadataModel(
            bias_5_3=sample_metrics.bias_5_3,
            duplicates=sample_metrics.pct_duplication,
            gc_content=sample_metrics.after_filtering_gc_content,
            input_amount=input_amount,
            insert_size=None,
            insert_size_peak=None,
            mapped_reads=sample_metrics.reads_aligned
            * 2
            / sample_metrics.before_filtering_total_reads,
            mean_length_r1=sample_metrics.after_filtering_read1_mean_length,
            million_read_pairs=get_million_read_pairs(
                reads=sample_metrics.before_filtering_total_reads
            ),
            mrna_bases=sample_metrics.pct_mrna_bases,
            pct_adapter=sample_metrics.pct_adapter,
            pct_surviving=sample_metrics.pct_surviving,
            q20_rate=sample_metrics.after_filtering_q20_rate,
            q30_rate=sample_metrics.after_filtering_q30_rate,
            ribosomal_bases=sample_metrics.pct_ribosomal_bases,
            rin=rin,
            uniquely_mapped_reads=sample_metrics.uniquely_mapped_percent,
        )

    def get_genome_build(self, analysis_metadata: AnalysisModel) -> str:
        """Return build version of the genome reference of a specific case."""
        return GenomeVersion.hg38.value

    def get_report_accreditation(
        self, samples: list[SampleModel], analysis_metadata: AnalysisModel
    ) -> bool:
        """Checks if the report is accredited or not. Rnafusion is not an accredited workflow."""
        return False

    def get_scout_uploaded_file_from_hk(self, case_id: str, scout_tag: str) -> str | None:
        """Return file path of the uploaded to Scout file given its tag."""
        return None

    def get_template_name(self) -> str:
        """Return template name to render the delivery report."""
        return Pipeline.RNAFUSION + "_report.html"

    def get_required_fields(self, case: CaseModel) -> dict:
        """Return dictionary with the delivery report required fields for Rnafusion."""
        return {
            "report": REQUIRED_REPORT_FIELDS,
            "customer": REQUIRED_CUSTOMER_FIELDS,
            "case": REQUIRED_CASE_FIELDS,
            "applications": self.get_application_required_fields(
                case=case, required_fields=REQUIRED_APPLICATION_FIELDS
            ),
            "data_analysis": REQUIRED_DATA_ANALYSIS_RNAFUSION_FIELDS,
            "samples": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_RNAFUSION_FIELDS
            ),
            "methods": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METHODS_FIELDS
            ),
            "timestamps": self.get_timestamp_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_TIMESTAMP_FIELDS
            ),
            "metadata": self.get_sample_required_fields(
                case=case, required_fields=REQUIRED_SAMPLE_METADATA_RNAFUSION_FIELDS
            ),
        }



"""PDC related constants"""
from enum import IntEnum

from cg.utils.enums import ListEnum


class DSMCParameters(ListEnum):
    ARCHIVE_COMMAND: list = ["archive"]
    QUERY_COMMAND: list = ["q", "archive"]
    RETRIEVE_COMMAND: list = ["retrieve", "-replace=yes"]



import logging
from pathlib import Path
from typing import Iterable

import click
from housekeeper.store.models import File

from cg.apps.crunchy.files import update_metadata_paths
from cg.cli.compress.helpers import update_compress_api
from cg.constants import SequencingFileTag
from cg.constants.constants import DRY_RUN
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Sample

LOG = logging.getLogger(__name__)


@click.command("sample")
@click.argument("sample-id", type=str)
@DRY_RUN
@click.pass_obj
def store_sample(context: CGConfig, sample_id: str, dry_run: bool) -> int:
    """Include links to decompressed FASTQ files belonging to this sample in Housekeeper."""
    compress_api: CompressAPI = context.meta_apis["compress_api"]
    status_db: Store = context.status_db
    update_compress_api(compress_api, dry_run=dry_run)
    sample = status_db.get_sample_by_internal_id(internal_id=sample_id)
    if not sample:
        LOG.warning(f"Could not find {sample_id}")
        return 0
    was_decompressed: bool = compress_api.add_decompressed_fastq(sample)
    if not was_decompressed:
        LOG.info(f"Skipping sample {sample_id}")
        return 0
    LOG.info(f"Stored fastq files for {sample_id}")
    return 1


def invoke_store_samples(context: click.Context, sample_ids: list[str], dry_run: bool) -> int:
    """Invoke store sample for sample ids."""
    stored_individuals: int = 0
    for sample_id in sample_ids:
        stored_count: int = context.invoke(store_sample, sample_id=sample_id, dry_run=dry_run)
        stored_individuals += stored_count
    return stored_individuals


@click.command("case")
@click.argument("case-id", type=str)
@DRY_RUN
@click.pass_context
def store_case(context: click.Context, case_id: str, dry_run: bool) -> None:
    """Include links to decompressed FASTQ files belonging to this case in Housekeeper."""

    status_db: Store = context.obj.status_db
    try:
        sample_ids: Iterable[str] = status_db.get_sample_ids_by_case_id(case_id=case_id)
        stored_individuals: int = invoke_store_samples(
            context=context, dry_run=dry_run, sample_ids=sample_ids
        )
    except CaseNotFoundError:
        return
    LOG.info(f"Stored fastq files for {stored_individuals} samples")


@click.command("flow-cell")
@click.argument("flow-cell-id", type=str)
@DRY_RUN
@click.pass_context
def store_flow_cell(context: click.Context, flow_cell_id: str, dry_run: bool) -> None:
    """Include links to decompressed FASTQ files belonging to this flow cell in Housekeeper."""

    status_db: Store = context.obj.status_db
    samples: list[Sample] = status_db.get_samples_from_flow_cell(flow_cell_id=flow_cell_id)
    stored_individuals: int = invoke_store_samples(
        context=context, dry_run=dry_run, sample_ids=[sample.internal_id for sample in samples]
    )
    LOG.info(f"Stored fastq files for {stored_individuals} samples")


@click.command("ticket")
@click.argument("ticket", type=str)
@DRY_RUN
@click.pass_context
def store_ticket(context: click.Context, ticket: str, dry_run: bool) -> None:
    """Include links to decompressed FASTQ files belonging to a ticket in Housekeeper."""
    status_db: Store = context.obj.status_db
    samples: list[Sample] = status_db.get_samples_from_ticket(ticket=ticket)
    stored_individuals: int = invoke_store_samples(
        context=context, dry_run=dry_run, sample_ids=[sample.internal_id for sample in samples]
    )
    LOG.info(f"Stored fastq files for {stored_individuals} samples")


@click.command("demultiplexed-flow-cell")
@click.argument("flow-cell-id", type=str)
@DRY_RUN
@click.pass_obj
def store_demultiplexed_flow_cell(context: click.Context, flow_cell_id: str, dry_run: bool) -> None:
    """Include all files in a flow cell bundle and its corresponding sequencing files. Updates SPRING metadata file"""
    compress_api: CompressAPI = context.meta_apis["compress_api"]
    status_db: Store = context.status_db
    update_compress_api(compress_api, dry_run=dry_run)

    samples: list[Sample] = status_db.get_samples_from_flow_cell(flow_cell_id=flow_cell_id)
    bundle_names: list[str] = (
        [sample.internal_id for sample in samples if sample.internal_id] + [flow_cell_id]
        if samples
        else [flow_cell_id]
    )
    for bundle_name in bundle_names:
        compress_api.hk_api.include_files_to_latest_version(bundle_name=bundle_name)

    if not samples:
        LOG.info(f"No samples found for flow cell {flow_cell_id} in the database")
        return
    for sample in samples:
        spring_metadata_files: list[File] = compress_api.hk_api.get_files_from_latest_version(
            bundle_name=sample.internal_id, tags=[SequencingFileTag.SPRING_METADATA]
        )
        if spring_metadata_files:
            for spring_metadata_file in spring_metadata_files:
                LOG.info(
                    f"Updating file paths in SPRING metadata file {spring_metadata_file.full_path}: for sample: {sample.internal_id}"
                )
                update_metadata_paths(
                    spring_metadata_path=spring_metadata_file.full_path,
                    new_parent_path=Path(spring_metadata_file.full_path).parent,
                )



"""Tests for the functions in the store api status related to the customer module."""


from cg.constants.invoice import CustomerNames
from cg.store import Store
from cg.store.models import Customer
from tests.store_helpers import StoreHelpers


def test_get_customers_to_invoice(
    store_with_samples_for_multiple_customers: Store,
    helpers: StoreHelpers,
    three_customer_ids: list[str],
):
    """Test that customers to invoice can be returned."""
    # GIVEN a database with samples for a customer

    # WHEN getting the customers to invoice
    customers: list[Customer] = store_with_samples_for_multiple_customers.get_customers_to_invoice(
        records=store_with_samples_for_multiple_customers.get_samples_to_invoice_query()
    )

    # THEN the customers should be returned
    assert customers
    assert len(customers) == 2
    for customer in customers:
        assert customer.internal_id in three_customer_ids

    # THEN the customers to invoice should not contain cust000
    for customer in customers:
        assert customer.internal_id != CustomerNames.CG_INTERNAL_CUSTOMER



"""Tests for the samples to deliver api"""

import datetime as dt

from cg.constants import DataDelivery
from cg.constants.constants import Pipeline
from cg.store.models import Case, Sample


def test_get_delivery_arguments(case: Case):
    """Testing the parsing of delivery arguments from the case data_delivery."""
    # GIVEN a DataDelivery
    case.data_delivery = DataDelivery.FASTQ_ANALYSIS_SCOUT

    # WHEN parsing the delivery types
    delivery_types: set[str] = case.get_delivery_arguments()

    # THEN the correct delivery types should be returned
    assert delivery_types == {Pipeline.MIP_DNA, Pipeline.FASTQ}


def test_list_samples_to_deliver(base_store, helpers):
    """Test to fetch samples ready for delivery"""
    store = base_store
    # GIVEN a populated store without samples
    assert len(store._get_query(table=Sample).all()) == 0
    # GIVEN inserting a sample that should be delivered
    helpers.add_sample(store, last_sequenced_at=dt.datetime.now())
    assert len(store._get_query(table=Sample).all()) == 1

    # WHEN asking for samples to deliver
    samples_to_deliver: list[Sample] = store.get_samples_to_deliver()
    # THEN it should return the sample which is ready to deliver
    assert len(samples_to_deliver) == 1
    assert isinstance(samples_to_deliver[0].last_sequenced_at, dt.datetime)


def test_list_samples_to_deliver_multiple_samples(base_store, helpers):
    """Test to fetch samples ready for delivery and avoid the ones that are not"""
    store = base_store
    # GIVEN a populated store with two samples where one is scheduled for delivery
    helpers.add_sample(store, last_sequenced_at=dt.datetime.now())
    helpers.add_sample(
        store,
        name="delivered",
        last_sequenced_at=dt.datetime.now(),
        delivered_at=dt.datetime.now(),
    )
    assert len(store._get_query(table=Sample).all()) == 2

    # WHEN asking for samples to deliver
    samples_to_deliver: list[Sample] = store.get_samples_to_deliver()
    # THEN it should return the sample which is ready to deliver
    assert len(samples_to_deliver) == 1
    assert isinstance(samples_to_deliver[0].last_sequenced_at, dt.datetime)



import click
from click.core import ParameterSource

from cg.constants.archiving import DEFAULT_SPRING_ARCHIVE_COUNT
from cg.meta.archive.archive import SpringArchiveAPI
from cg.models.cg_config import CGConfig


@click.group()
def archive():
    """Archive utilities."""
    pass


@archive.command("archive-spring-files")
@click.option(
    "-l",
    "--limit",
    help="Give a limit to the amount of files to archive.",
    default=DEFAULT_SPRING_ARCHIVE_COUNT,
    show_default=True,
)
@click.option(
    "--archive-all",
    help="Use in order to archive all non-archived-files",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.pass_obj
def archive_spring_files(context: CGConfig, limit: int | None, archive_all: bool):
    """Archives non-archived spring files.
    Raises:
        click.Abort if both a limit to the number of spring files to archive and archive_all is specified.
    """

    if (
        click.get_current_context().get_parameter_source("limit") == ParameterSource.COMMANDLINE
        and limit
        and archive_all
    ):
        click.echo(
            "Incorrect input parameters - please do not provide both a limit and set --archive-all."
        )
        raise click.Abort

    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow_config,
    )
    spring_archive_api.archive_spring_files_and_add_archives_to_housekeeper(
        spring_file_count_limit=None if archive_all else limit
    )


@archive.command("update-job-statuses")
@click.pass_obj
def update_job_statuses(context: CGConfig):
    """Queries ongoing jobs and updates Housekeeper."""
    spring_archive_api = SpringArchiveAPI(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        data_flow_config=context.data_flow_config,
    )
    spring_archive_api.update_statuses_for_ongoing_tasks()



from pydantic.v1 import BaseModel, validator
from typing_extensions import Literal

from cg.constants import Priority

SEX_MAP = {"male": "M", "female": "F"}


class Udf(BaseModel):
    application: str
    capture_kit: str | None
    collection_date: str | None
    comment: str | None
    concentration: str | None
    concentration_sample: str | None
    customer: str
    control: str | None
    data_analysis: str | None
    data_delivery: str | None
    elution_buffer: str | None
    extraction_method: str | None
    family_name: str | None = "NA"
    formalin_fixation_time: str | None
    index: str | None
    index_number: str | None
    lab_code: str | None
    organism: str | None
    organism_other: str | None
    original_lab: str | None
    original_lab_address: str | None
    pool: str | None
    post_formalin_fixation_time: str | None
    pre_processing_method: str | None
    primer: str | None
    priority: str = Priority.standard.name
    quantity: str | None
    reference_genome: str | None
    region: str | None
    region_code: str | None
    require_qc_ok: bool = False
    rml_plate_name: str | None
    selection_criteria: str | None
    sex: Literal["M", "F", "unknown"] = "unknown"
    skip_reception_control: bool | None = None
    source: str = "NA"
    tissue_block_size: str | None
    tumour: bool | None = False
    tumour_purity: str | None
    volume: str | None
    well_position_rml: str | None
    verified_organism: bool | None

    @validator("sex", pre=True)
    def validate_sex(cls, value: str):
        return SEX_MAP.get(value, "unknown")


class LimsSample(BaseModel):
    name: str
    container: str = "Tube"
    container_name: str | None
    well_position: str | None
    index_sequence: str | None
    udfs: Udf | None

    @classmethod
    def parse_obj(cls, obj: dict):
        parsed_obj: LimsSample = super().parse_obj(obj)
        udf: Udf = Udf.parse_obj(obj)
        parsed_obj.udfs = udf
        return parsed_obj



from cg.store.api.core import Store
from cg.store.filters.status_user_filters import filter_user_by_email
from cg.store.models import User


def test_filter_user_by_email_returns_correct_user(store_with_users: Store):
    """Test getting user by email."""

    # GIVEN a store with a user
    user: User = store_with_users._get_query(table=User).first()
    assert user

    # WHEN retrieving the user by email
    filtered_user: User = filter_user_by_email(
        users=store_with_users._get_query(table=User),
        email=user.email,
    ).first()

    # THEN a user should be returned
    assert isinstance(filtered_user, User)

    # THEN the email should match
    assert filtered_user.email == user.email


def test_filter_user_by_email_returns_none_for_nonexisting_email(
    store_with_users: Store, non_existent_email: str
):
    """Test getting user by email when the email does not exist."""

    # GIVEN a non-existing email

    # WHEN retrieving the user by email
    filtered_user: User = filter_user_by_email(
        users=store_with_users._get_query(table=User),
        email=non_existent_email,
    ).first()

    # THEN no user should be returned
    assert filtered_user is None


def test_filter_user_by_email_none_returns_none(store_with_users: Store):
    """Test getting user by email None."""

    # WHEN retrieving the user by email None
    filtered_user: User = filter_user_by_email(
        users=store_with_users._get_query(table=User),
        email=None,
    ).first()

    # THEN no user should be returned
    assert filtered_user is None



import logging
import os
from pathlib import Path

import click
from pydantic import ValidationError

from cg.apps.demultiplex.sample_sheet.create import create_sample_sheet
from cg.apps.demultiplex.sample_sheet.models import FlowCellSample
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_sample_sheet_from_file
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims.sample_sheet import get_flow_cell_samples
from cg.constants.constants import DRY_RUN, FileFormat
from cg.constants.demultiplexing import OPTION_BCL_CONVERTER
from cg.exc import FlowCellError, HousekeeperFileMissingError, SampleSheetError
from cg.io.controller import WriteFile, WriteStream
from cg.meta.demultiplex.housekeeper_storage_functions import add_sample_sheet_path_to_housekeeper
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData

LOG = logging.getLogger(__name__)


@click.group(name="samplesheet")
def sample_sheet_commands():
    """Command group for the sample sheet commands."""


@sample_sheet_commands.command(name="validate")
@click.argument("sheet", type=click.Path(exists=True, dir_okay=False))
def validate_sample_sheet(sheet: click.Path):
    """Validate a sample sheet."""
    LOG.info(f"Validating {sheet} sample sheet")
    try:
        get_sample_sheet_from_file(Path(sheet))
    except ValidationError as error:
        LOG.warning(error)
        raise click.Abort from error
    LOG.info("Sample sheet passed validation")


@sample_sheet_commands.command(name="create")
@click.argument("flow-cell-name")
@OPTION_BCL_CONVERTER
@DRY_RUN
@click.option("--force", is_flag=True, help="Skips the validation of the sample sheet")
@click.pass_obj
def create_sheet(
    context: CGConfig, flow_cell_name: str, bcl_converter: str, dry_run: bool, force: bool = False
):
    """Create a sample sheet or hard-link it from Housekeeper in the flow cell directory.

    flow-cell-name is the flow cell run directory name, e.g. '201203_D00483_0200_AHVKJCDRXX'

    Search the flow cell in the directory specified in config.
    """
    LOG.info(f"Creating sample sheet for flow cell {flow_cell_name}")
    hk_api: HousekeeperAPI = context.housekeeper_api
    flow_cell_path: Path = Path(context.flow_cells_dir, flow_cell_name)
    if not flow_cell_path.exists():
        LOG.warning(f"Could not find flow cell {flow_cell_path}")
        raise click.Abort
    try:
        flow_cell = FlowCellDirectoryData(
            flow_cell_path=flow_cell_path, bcl_converter=bcl_converter
        )
    except FlowCellError as error:
        raise click.Abort from error
    flow_cell_id: str = flow_cell.id

    try:
        sample_sheet_path: Path = hk_api.get_sample_sheet_path(flow_cell_id)
    except HousekeeperFileMissingError:
        sample_sheet_path = None

    if flow_cell.sample_sheet_exists():
        LOG.info("Sample sheet already exists in flow cell directory")
        if not dry_run:
            add_sample_sheet_path_to_housekeeper(
                flow_cell_directory=flow_cell_path, flow_cell_name=flow_cell_id, hk_api=hk_api
            )
        return
    elif sample_sheet_path:
        LOG.debug(
            "Sample sheet already exists in Housekeeper. Hard-linking it to flow cell directory"
        )
        if not dry_run:
            os.link(src=sample_sheet_path, dst=flow_cell.sample_sheet_path)
        return
    lims_samples: list[FlowCellSample] = list(
        get_flow_cell_samples(
            lims=context.lims_api,
            flow_cell_id=flow_cell_id,
            flow_cell_sample_type=flow_cell.sample_type,
        )
    )
    if not lims_samples:
        LOG.warning(f"Could not find any samples in LIMS for {flow_cell_id}")
        raise click.Abort
    try:
        sample_sheet_content: list[list[str]] = create_sample_sheet(
            flow_cell=flow_cell,
            lims_samples=lims_samples,
            force=force,
        )
    except (FileNotFoundError, FileExistsError, FlowCellError) as error:
        raise click.Abort from error

    if dry_run:
        click.echo(
            WriteStream.write_stream_from_content(
                file_format=FileFormat.CSV, content=sample_sheet_content
            )
        )
        return
    LOG.info(f"Writing sample sheet to {flow_cell.sample_sheet_path.resolve()}")
    WriteFile.write_file_from_content(
        content=sample_sheet_content,
        file_format=FileFormat.CSV,
        file_path=flow_cell.sample_sheet_path,
    )
    LOG.info("Adding sample sheet to Housekeeper")
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=flow_cell_path, flow_cell_name=flow_cell_id, hk_api=hk_api
    )


@sample_sheet_commands.command(name="create-all")
@DRY_RUN
@click.pass_obj
def create_all_sheets(context: CGConfig, dry_run: bool):
    """Command to create sample sheets for all flow cells that lack a sample sheet.

    Search flow cell directories for run parameters and create a sample sheets based on the
    information.
    """
    hk_api: HousekeeperAPI = context.housekeeper_api
    flow_cell_runs_dir: Path = Path(context.flow_cells_dir)
    for sub_dir in flow_cell_runs_dir.iterdir():
        if not sub_dir.is_dir():
            continue
        LOG.debug(f"Found directory {sub_dir}")
        try:
            flow_cell = FlowCellDirectoryData(flow_cell_path=sub_dir)
        except FlowCellError:
            continue
        flow_cell_id: str = flow_cell.id

        try:
            sample_sheet_path: Path | None = hk_api.get_sample_sheet_path(flow_cell_id)
        except HousekeeperFileMissingError:
            sample_sheet_path = None

        if flow_cell.sample_sheet_exists():
            LOG.debug("Sample sheet already exists in flow cell directory")
            if not dry_run:
                add_sample_sheet_path_to_housekeeper(
                    flow_cell_directory=sub_dir, flow_cell_name=flow_cell_id, hk_api=hk_api
                )
            continue
        elif sample_sheet_path:
            LOG.debug(
                "Sample sheet already exists in Housekeeper. Copying it to flow cell directory"
            )
            if not dry_run:
                os.link(src=sample_sheet_path, dst=flow_cell.sample_sheet_path)
            continue
        LOG.info(f"Creating sample sheet for flow cell {flow_cell.id}")
        lims_samples: list[FlowCellSample] = list(
            get_flow_cell_samples(
                lims=context.lims_api,
                flow_cell_id=flow_cell_id,
                flow_cell_sample_type=flow_cell.sample_type,
            )
        )
        if not lims_samples:
            LOG.warning(f"Could not find any samples in LIMS for {flow_cell_id}")
            continue

        try:
            sample_sheet_content: list[list[str]] = create_sample_sheet(
                flow_cell=flow_cell,
                lims_samples=lims_samples,
            )
        except (
            FileNotFoundError,
            FileExistsError,
            ValidationError,
            FlowCellError,
            SampleSheetError,
        ) as exc:
            LOG.warning(f"Sample sheet generation failed for flow cell {flow_cell_id} due to {exc}")
            continue

        if dry_run:
            click.echo(
                WriteStream.write_stream_from_content(
                    file_format=FileFormat.CSV, content=sample_sheet_content
                )
            )
            continue
        LOG.info(f"Writing sample sheet to {flow_cell.sample_sheet_path.resolve()}")
        WriteFile.write_file_from_content(
            content=sample_sheet_content,
            file_format=FileFormat.CSV,
            file_path=flow_cell.sample_sheet_path,
        )
        LOG.info("Adding sample sheet to Housekeeper")
        add_sample_sheet_path_to_housekeeper(
            flow_cell_directory=sub_dir, flow_cell_name=flow_cell_id, hk_api=hk_api
        )



from enum import StrEnum

DEFAULT_SPRING_ARCHIVE_COUNT = 200


class ArchiveLocations(StrEnum):
    """Archive locations for the different customers' Spring files."""

    KAROLINSKA_BUCKET: str = "karolinska_bucket"


PDC_ARCHIVE_LOCATION: str = "PDC"



"""Import all constants here for easy access"""

from cg.constants.compression import (
    FASTQ_DELTA,
    FASTQ_FIRST_READ_SUFFIX,
    FASTQ_SECOND_READ_SUFFIX,
)
from cg.constants.constants import (
    CAPTUREKIT_CANCER_OPTIONS,
    CAPTUREKIT_OPTIONS,
    COLLABORATORS,
    COMBOS,
    CONTAINER_OPTIONS,
    DEFAULT_CAPTURE_KIT,
    PREP_CATEGORIES,
    SEX_OPTIONS,
    STATUS_OPTIONS,
    DataDelivery,
    FileExtensions,
    FlowCellStatus,
)
from cg.constants.gene_panel import GenePanelMasterList
from cg.constants.housekeeper_tags import (
    HK_FASTQ_TAGS,
    HK_MULTIQC_HTML_TAG,
    SequencingFileTag,
)
from cg.constants.paths import TMP_DIR
from cg.constants.priority import Priority
from cg.constants.process import EXIT_FAIL, EXIT_SUCCESS
from cg.constants.report import *
from cg.constants.sample_sources import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.constants.sequencing import FLOWCELL_Q30_THRESHOLD
from cg.constants.symbols import SINGLE_QUOTE, SPACE



"""This script tests the cli methods to set values of a case in status-db."""
from click.testing import CliRunner

from cg.cli.set.case import set_case
from cg.constants import EXIT_SUCCESS, DataDelivery, Pipeline
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Case
from tests.store_helpers import StoreHelpers


def test_set_case_without_options(
    cli_runner: CliRunner,
    base_context: CGConfig,
    base_store: Store,
    helpers: StoreHelpers,
):
    """Test to set a case using only the required arguments."""
    # GIVEN a database with a case
    case_id: str = helpers.add_case(store=base_store).internal_id
    assert base_store._get_query(table=Case).count() == 1

    # WHEN setting a case
    result = cli_runner.invoke(set_case, [case_id], obj=base_context)

    # THEN it should abort
    assert result.exit_code != EXIT_SUCCESS


def test_set_case_bad_family(
    cli_runner: CliRunner, base_context: CGConfig, case_id_does_not_exist: str
):
    """Test to set a case using a non-existing case id."""
    # GIVEN an empty database

    # WHEN setting a case
    result = cli_runner.invoke(set_case, [case_id_does_not_exist], obj=base_context)

    # THEN it should complain on missing case
    assert result.exit_code != EXIT_SUCCESS


def test_set_case_bad_panel(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test to set a case using a non-existing panel."""
    # GIVEN a database with a case

    # WHEN setting a case
    panel_id: str = "dummy_panel"
    case_id: str = helpers.add_case(store=base_store).internal_id
    result = cli_runner.invoke(set_case, [case_id, "--panel", panel_id], obj=base_context)

    # THEN it should complain in missing panel instead of setting a value
    assert result.exit_code != EXIT_SUCCESS
    assert panel_id not in base_store._get_query(table=Case).first().panels


def test_set_case_panel(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test to set a case using an existing panel."""
    # GIVEN a database with a case and a panel not yet added to the case
    panel_id: str = helpers.ensure_panel(store=base_store, panel_abbreviation="a_panel").name
    case_id: str = helpers.add_case(store=base_store).internal_id
    case_query = base_store._get_query(table=Case)

    assert panel_id not in case_query.first().panels

    # WHEN setting a panel of a case
    result = cli_runner.invoke(set_case, [case_id, "--panel", panel_id], obj=base_context)

    # THEN it should set panel on the case
    assert result.exit_code == EXIT_SUCCESS
    assert panel_id in case_query.first().panels


def test_set_case_priority(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test that the added case gets the priority we send in."""
    # GIVEN a database with a case
    case_id: str = helpers.add_case(base_store).internal_id
    priority: str = "priority"
    case_query = base_store._get_query(table=Case)

    assert case_query.first().priority_human != priority

    # WHEN setting a case
    result = cli_runner.invoke(
        set_case, [case_id, "--priority", priority], obj=base_context, catch_exceptions=False
    )

    # THEN it should have been set
    assert result.exit_code == EXIT_SUCCESS
    assert case_query.count() == 1
    assert case_query.first().priority_human == priority


def test_set_case_customer(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test to set a case using an existing customer."""
    # GIVEN a database with a case and a customer not yet on the case
    customer_id: str = helpers.ensure_customer(
        store=base_store, customer_id="a_customer"
    ).internal_id
    case_to_alter: Case = helpers.add_case(store=base_store)
    assert customer_id != case_to_alter.customer.internal_id

    # WHEN setting a customer of a case
    result = cli_runner.invoke(
        set_case, [case_to_alter.internal_id, "--customer-id", customer_id], obj=base_context
    )

    # THEN it should set customer on the case
    assert result.exit_code == EXIT_SUCCESS
    assert customer_id == case_to_alter.customer.internal_id


def test_set_case_bad_data_analysis(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test to set a case using a non-existing data_analysis."""
    # GIVEN a database with a case

    # WHEN setting a data_analysis on a case
    data_analysis: str = "dummy_pipeline"
    case_id: str = helpers.add_case(store=base_store).internal_id
    result = cli_runner.invoke(
        set_case, [case_id, "--data-analysis", data_analysis], obj=base_context
    )

    # THEN it should complain in invalid data_analysis instead of setting a value
    assert result.exit_code != EXIT_SUCCESS
    assert str(data_analysis) != base_store._get_query(table=Case).first().data_analysis


def test_set_case_data_analysis(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test to set a case using an existing data_analysis."""

    # GIVEN a database with a case and a data_analysis not yet set on the case
    data_analysis: str = Pipeline.FASTQ
    case_to_alter: str = helpers.add_case(base_store)
    assert str(data_analysis) != case_to_alter.data_analysis

    # WHEN setting a data_analysis of a case
    result = cli_runner.invoke(
        set_case,
        [case_to_alter.internal_id, "--data-analysis", str(data_analysis)],
        obj=base_context,
    )

    # THEN it should set data_analysis on the case
    assert result.exit_code == EXIT_SUCCESS
    assert str(data_analysis) == case_to_alter.data_analysis


def test_set_case_bad_data_delivery(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test to set a case using a non-existing data_delivery."""
    # GIVEN a database with a case

    # WHEN setting a data_delivery on a case
    data_delivery: str = "dummy_delivery"
    case_id: str = helpers.add_case(base_store).internal_id
    result = cli_runner.invoke(
        set_case, [case_id, "--data-delivery", data_delivery], obj=base_context
    )

    # THEN it should complain in invalid data_delivery instead of setting a value
    assert result.exit_code != EXIT_SUCCESS
    assert str(data_delivery) != base_store._get_query(table=Case).first().data_delivery


def test_set_case_data_delivery(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers: StoreHelpers
):
    """Test to set a case using an existing data_delivery."""

    # GIVEN a database with a case and a data_delivery not yet set on the case
    data_delivery: str = DataDelivery.FASTQ
    case_to_alter: str = helpers.add_case(base_store)
    assert str(data_delivery) != case_to_alter.data_delivery

    # WHEN setting a data_delivery of a case
    result = cli_runner.invoke(
        set_case,
        [case_to_alter.internal_id, "--data-delivery", str(data_delivery)],
        obj=base_context,
    )

    # THEN it should set data_delivery on the case
    assert result.exit_code == EXIT_SUCCESS
    assert str(data_delivery) == case_to_alter.data_delivery



"""Module to create API requests."""
import requests
from requests import Response


def put(url: str, headers: dict, json: dict, verify: bool = True) -> Response:
    """Create PUT request."""
    return requests.put(url=url, headers=headers, json=json, verify=verify)


def post(url: str, headers: dict, json: dict, verify: bool = True) -> Response:
    """Create POST request."""
    return requests.post(url=url, headers=headers, json=json, verify=verify)


def delete(url: str, headers: dict, json: dict, verify: bool = True) -> Response:
    """Create DELETE request."""
    return requests.delete(url=url, headers=headers, json=json, verify=verify)


def get(url: str, headers: dict, json: dict, verify: bool = True) -> Response:
    """Create GET request."""
    return requests.get(url=url, headers=headers, json=json, verify=verify)


def patch(url: str, headers: dict, json: dict, verify: bool = True) -> Response:
    """Create PATCH request."""
    return requests.patch(url=url, headers=headers, json=json, verify=verify)



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.store.models import Collaboration


def filter_collaboration_by_internal_id(collaborations: Query, internal_id: str, **kwargs) -> Query:
    """Return collaboration by internal_id."""
    return collaborations.filter(Collaboration.internal_id == internal_id)


class CollaborationFilter(Enum):
    """Define Collaboration filter functions."""

    FILTER_BY_INTERNAL_ID: Callable = filter_collaboration_by_internal_id


def apply_collaboration_filter(
    collaborations: Query, filter_functions: list[Callable], internal_id: str | None = None
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        collaborations: Query = function(
            collaborations=collaborations,
            internal_id=internal_id,
        )
    return collaborations



"""Tests for the clean flow cells API."""
import logging
import time

import mock
import pytest
from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.time import TWENTY_ONE_DAYS_IN_SECONDS
from cg.exc import CleanFlowCellFailedError, HousekeeperFileMissingError
from cg.meta.clean.clean_flow_cells import CleanFlowCellAPI
from cg.store import Store
from cg.store.models import Flowcell, SampleLaneSequencingMetrics


def test_get_flow_cell_from_statusdb(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to get a flow cell from statusdb."""

    # GIVEN a clean flow cell api with a store that contains a flow cell

    # WHEN retrieving the flow cell from the store
    flow_cell: Flowcell = flow_cell_clean_api_can_be_removed.get_flow_cell_from_status_db()

    # THEN a flow cell is retrieved
    assert isinstance(flow_cell, Flowcell)


def test_is_flow_cell_in_statusdb(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to check that a flow cell is in statusdb."""
    # GIVEN a clean flow cell api with a store that contains a flow cell

    # WHEN checking whether a flow cell is in status DB
    is_in_statusdb: bool = flow_cell_clean_api_can_be_removed.is_flow_cell_in_statusdb()

    # THEN checking whether a flow cell is in statusdb is TRUE
    assert is_in_statusdb


def test_is_flow_cell_backed_up(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to check that a flow cell has been backed up."""
    # GIVEN a clean flow cell api with a flow cell that has been backed up

    # WHEN checking whether the flow cell has been backed
    is_backed_up: bool = flow_cell_clean_api_can_be_removed.is_flow_cell_backed_up()

    # THEN checking whether the flow cell has been backed up is TRUE
    assert is_backed_up


def test_is_not_flow_cell_backed_up(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to check that a flow cell has been backed up."""
    # GIVEN a clean flow cell api with a flow cell that has not been backed up
    flow_cell_clean_api_can_be_removed.get_flow_cell_from_status_db().has_backup = False

    # WHEN checking whether the flow cell has been backed
    is_backed_up: bool = flow_cell_clean_api_can_be_removed.is_flow_cell_backed_up()

    # THEN checking whether the flow cell has been backed up is FALSE
    assert not is_backed_up


def test_get_sequencing_metrics_for_flow_cell_from_statusdb(
    flow_cell_clean_api_can_be_removed: CleanFlowCellAPI,
):
    """Test to get a sequencing metrics for a flow cell from statusdb."""

    # GIVEN a clean flow cell api with a store that contains a flow cell with SampleLaneSequencingMetrics entries

    # WHEN retrieving the flow cell from the store
    metrics: list[
        SampleLaneSequencingMetrics
    ] = flow_cell_clean_api_can_be_removed.get_sequencing_metrics_for_flow_cell()

    # THEN a SampleLaneSequencingMetrics entry is retrieved
    assert isinstance(metrics[0], SampleLaneSequencingMetrics)


def test_has_sequencing_metrics_in_statusdb(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to check that a flow cell has SampleLaneSequencingMetrics entries in statusdb."""

    # GIVEN a clean flow cell api with a store that contains a flow cell with SampleLaneSequencingMetrics entries

    # WHEN THEN checking whether a flow cell has SampleLaneSequencingMetrics entries'
    has_metrics: bool = flow_cell_clean_api_can_be_removed.has_sequencing_metrics_in_statusdb()

    # THEN checking whether a flow cell has SampleLaneSequencingMetrics entries returns True
    assert has_metrics


def test_is_directory_older_than_21_days_pass(
    flow_cell_clean_api_can_be_removed: CleanFlowCellAPI,
):
    """Test to check whether a directory is older than 21 days."""

    # GIVEN a clean flow cell api with a flow cell that can be deleted

    # WHEN checking whether a given flow cell directory is older than 21 days

    with mock.patch(
        "time.time",
        return_value=time.time() + TWENTY_ONE_DAYS_IN_SECONDS,
    ):
        is_older_that_21_days: bool = (
            flow_cell_clean_api_can_be_removed.is_directory_older_than_21_days()
        )

    # THEN checking whether a given flow cell directory is older than 21 days is TRUE
    assert is_older_that_21_days


def test_is_directory_older_than_21_days_fail(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to check whether a directory is older than 21 days."""

    # GIVEN a clean flow cell api with a current time that is set to now.

    # WHEN checking whether a given flow cell directory is older than 21 days
    is_older_than_21_days: bool = (
        flow_cell_clean_api_can_be_removed.is_directory_older_than_21_days()
    )

    # THEN checking whether a given flow cell directory is older than 21 days is FALSE
    assert not is_older_than_21_days


def test_has_sample_sheet_in_housekeeper(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to check whether a flow cell has a sample sheet in housekeeper."""

    # GIVEN a clean flow cell api with a flow cell that has a sample sheet in housekeeper

    # WHEN checking whether the flow cell has a sample sheet in housekeeper
    has_sample_sheet: bool = flow_cell_clean_api_can_be_removed.has_sample_sheet_in_housekeeper()

    # THEN checking whether the flow cell has a sample sheet in housekeeper returns TRUE
    assert has_sample_sheet


def test_get_files_for_flow_cell_bundle(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test to get files for a flow cell bundle from housekeeper."""

    # GIVEN a clean flow cell api with a flow cell that has files in housekeeper

    # WHEN getting fastq, spring and metadata files that are tagged with the flow cell
    for tag in [
        SequencingFileTag.FASTQ,
        SequencingFileTag.SPRING,
        SequencingFileTag.SPRING_METADATA,
    ]:
        files: list[
            File
        ] = flow_cell_clean_api_can_be_removed.get_files_for_samples_on_flow_cell_with_tag(tag=tag)

        # THEN fastq, spring and spring metadata files are returned
        assert files


def test_can_flow_cell_be_deleted(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test the flow cell can be deleted check."""
    # GIVEN a flow cell that can be deleted

    with mock.patch(
        "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.is_directory_older_than_21_days",
        return_value=True,
    ):
        # WHEN checking that the flow cell can be deleted
        can_be_deleted: bool = (
            flow_cell_clean_api_can_be_removed.can_flow_cell_directory_be_deleted()
        )

    # THEN the check whether the flow cell can be deleted returns True
    assert can_be_deleted


def test_can_flow_cell_be_deleted_no_spring_with_fastq(
    flow_cell_clean_api_can_be_removed: CleanFlowCellAPI,
):
    """Test that a flow cell can be deleted when it has fastq files but no spring files."""
    # GIVEN a flow cell that can be deleted

    with mock.patch(
        "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.is_directory_older_than_21_days",
        return_value=True,
    ):
        with mock.patch(
            "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.has_spring_meta_data_files_for_samples_in_housekeeper",
            return_value=False,
        ):
            # WHEN checking that the flow cell can be deleted
            can_be_deleted: bool = (
                flow_cell_clean_api_can_be_removed.can_flow_cell_directory_be_deleted()
            )

    # THEN the check whether the flow cell can be deleted returns True
    assert can_be_deleted


def test_can_flow_cell_be_deleted_spring_no_fastq(
    flow_cell_clean_api_can_be_removed: CleanFlowCellAPI,
):
    """Test that a flow cell can be deleted when it has spring files but no fastq files."""
    # GIVEN a flow cell that can be deleted

    with mock.patch(
        "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.is_directory_older_than_21_days",
        return_value=True,
    ):
        with mock.patch(
            "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.has_fastq_files_for_samples_in_housekeeper",
            return_value=False,
        ):
            # WHEN checking that the flow cell can be deleted
            can_be_deleted: bool = (
                flow_cell_clean_api_can_be_removed.can_flow_cell_directory_be_deleted()
            )

    # THEN the check whether the flow cell can be deleted returns True
    assert can_be_deleted


def test_can_flow_cell_be_deleted_no_spring_no_fastq(
    flow_cell_clean_api_can_be_removed: CleanFlowCellAPI,
):
    """Test that a flow cell can not be deleted when it has no spring files and no fastq files."""
    # GIVEN a flow cell that can be deleted

    with mock.patch(
        "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.is_directory_older_than_21_days",
        return_value=True,
    ):
        with mock.patch(
            "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.has_fastq_files_for_samples_in_housekeeper",
            return_value=False,
        ):
            with mock.patch(
                "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.has_spring_meta_data_files_for_samples_in_housekeeper",
                return_value=False,
            ):
                # WHEN checking that the flow cell can be deleted

                # THEN a HousekeeperFileMissingError is raised
                with pytest.raises(HousekeeperFileMissingError):
                    flow_cell_clean_api_can_be_removed.can_flow_cell_directory_be_deleted()


def test_delete_flow_cell_directory(flow_cell_clean_api_can_be_removed: CleanFlowCellAPI):
    """Test that a flow cell directory is removed."""
    # GIVEN a flow cell that can be removed

    # GIVEN that the flow cell directory exists
    assert flow_cell_clean_api_can_be_removed.flow_cell.path.exists()

    # WHEN removing the flow cell directory
    with mock.patch(
        "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.is_directory_older_than_21_days",
        return_value=True,
    ):
        flow_cell_clean_api_can_be_removed.delete_flow_cell_directory()

    # THEN the flow cell directory is removed
    assert not flow_cell_clean_api_can_be_removed.flow_cell.path.exists()


def test_delete_flow_cell_directory_can_not_be_deleted(
    flow_cell_clean_api_can_not_be_removed: CleanFlowCellAPI,
):
    """Test delete a flow cell that does not pass all checks."""
    # GIVEN a flow cell that should not be removed.

    # GIVEN that the flow cell directory exists
    assert flow_cell_clean_api_can_not_be_removed.flow_cell.path.exists()

    # WHEN trying to remove the flow cell
    with pytest.raises(CleanFlowCellFailedError):
        flow_cell_clean_api_can_not_be_removed.delete_flow_cell_directory()

    # THEN the flow cell directory still exists
    flow_cell_clean_api_can_not_be_removed.flow_cell.path.exists()


def test_get_flow_cell_from_statusdb_does_not_exist(
    flow_cell_clean_api_can_not_be_removed: CleanFlowCellAPI,
):
    """Test retrieving a flow cell from statusDB that does not exist."""
    # GIVEN a CleanFlowCellAPI with a flow cell that is not in statusDB
    flow_cell_clean_api_can_not_be_removed.flow_cell.id = "flow_cell_does_not_exist"

    assert not flow_cell_clean_api_can_not_be_removed.status_db.get_flow_cell_by_name(
        flow_cell_clean_api_can_not_be_removed.flow_cell.id
    )

    # WHEN retrieving the flow cell from statusDB

    # THEN a ValueError is raised
    with pytest.raises(ValueError):
        flow_cell_clean_api_can_not_be_removed.get_flow_cell_from_status_db()


def test_flow_cell_has_not_fastq_files_in_housekeeper(
    flow_cell_clean_api_can_not_be_removed: CleanFlowCellAPI,
    store_with_flow_cell_not_to_clean: Store,
    housekeeper_api_with_flow_cell_not_to_clean: HousekeeperAPI,
    caplog,
):
    """Test to check wether a flow cell has files in Housekeeper."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a flow cell that has entries in StatusDB but does not have any files in Housekeeper for its samples
    flow_cell_clean_api_can_not_be_removed.status_db = store_with_flow_cell_not_to_clean
    flow_cell_clean_api_can_not_be_removed.hk_api = housekeeper_api_with_flow_cell_not_to_clean

    # WHEN checking whether a sample on a flow cell has fastq files

    # THEN the flow cell has no fastq files in housekeeper.
    assert not flow_cell_clean_api_can_not_be_removed.has_fastq_files_for_samples_in_housekeeper()
    assert SequencingFileTag.FASTQ in caplog.text


def test_flow_cell_has_not_spring_files_in_housekeeper(
    flow_cell_clean_api_can_not_be_removed: CleanFlowCellAPI,
    store_with_flow_cell_not_to_clean: Store,
    housekeeper_api_with_flow_cell_not_to_clean: HousekeeperAPI,
    caplog,
):
    """Test to check whether a flow cell has files in Housekeeper."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a flow cell that has entries in StatusDB but does not have any files in Housekeeper for its samples
    flow_cell_clean_api_can_not_be_removed.status_db = store_with_flow_cell_not_to_clean
    flow_cell_clean_api_can_not_be_removed.hk_api = housekeeper_api_with_flow_cell_not_to_clean

    # WHEN checking whether a sample on a flow cell has spring files

    # THEN the flow cell has no spring files for samples in housekeeper
    assert not flow_cell_clean_api_can_not_be_removed.has_spring_files_for_samples_in_housekeeper()
    assert SequencingFileTag.SPRING in caplog.text


def test_flow_cell_has_not_spring_meta_data_files_in_housekeeper(
    flow_cell_clean_api_can_not_be_removed: CleanFlowCellAPI,
    store_with_flow_cell_not_to_clean: Store,
    housekeeper_api_with_flow_cell_not_to_clean: HousekeeperAPI,
    caplog,
):
    """Test to check whether a flow cell has files in Housekeeper."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a flow cell that has entries in StatusDB but does not have any files in Housekeeper for its samples
    flow_cell_clean_api_can_not_be_removed.status_db = store_with_flow_cell_not_to_clean
    flow_cell_clean_api_can_not_be_removed.hk_api = housekeeper_api_with_flow_cell_not_to_clean

    # WHEN checking whether a sample on a flow cell has spring files

    # THEN the flow cell has no spring metadata files in housekeeper

    assert (
        not flow_cell_clean_api_can_not_be_removed.has_spring_meta_data_files_for_samples_in_housekeeper()
    )
    assert SequencingFileTag.SPRING_METADATA in caplog.text


def test_has_no_sample_files_in_housekeeper(
    flow_cell_clean_api_can_not_be_removed: CleanFlowCellAPI,
    store_with_flow_cell_not_to_clean: Store,
    housekeeper_api_with_flow_cell_not_to_clean: HousekeeperAPI,
):
    # GIVEN a flow cell that has entries in StatusDB but does not have any files in Housekeeper for its samples
    flow_cell_clean_api_can_not_be_removed.status_db = store_with_flow_cell_not_to_clean
    flow_cell_clean_api_can_not_be_removed.hk_api = housekeeper_api_with_flow_cell_not_to_clean

    # WHEN checking whether a sample on a flow cell has files in housekeeper

    # THEN a HousekeeperFileMissingError is raised
    with pytest.raises(HousekeeperFileMissingError):
        flow_cell_clean_api_can_not_be_removed.has_sample_fastq_or_spring_files_in_housekeeper()
